/**
 * Mar 28: Try method of calculating harmonic average of offers, and making delta epsilon*harmonic average
 * Mar 30: Try (again) method of updating Di when adding delta (not going over all bids)
 * Mar 31: again - this time add correction every 50 rounds - calculating Di from 0
 *         the problem files are prob.2.200.500.5 and prob.2.300.500.6
 * Apr 02: remove CopyProblem and CopyBid
 */
#include "Algorithm.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

double *Xi = NULL;
double *Yj = NULL;
double *alphaj = NULL;
double alpha = 0;
int B;
double *Di = NULL;
double maxD; // for dividing Yj
double delta = 0;
double sum_x; // the sum of x before normalizing
double sum_y; // the sum of y before normalizing
double minDelta;
double thresh; // for poor bids

void calcD (Problem *problem) {
	// first - erase previous values of Di
	for (int i=0; i<problem->num_of_items; i++)
		Di[i] = 0;
	// fill out Di
	for (int j=0; j<problem->num_of_bids; j++)
		for (int i=0; i<problem->bids[j]->items_in_bid; i++)
			Di[problem->bids[j]->items[i]] += Yj[j];
}		

// this assumes that Di is current - so only pick arg(maxD)
// side effect - update maxD
int calcB (Problem *problem) {
	int ans;
	// calculate maxD
	maxD = 0;
	for (int i=0; i<problem->num_of_items; i++)
		if (maxD < Di[i]) {
			maxD = Di[i];
			ans = i;
		}
	return ans;
}		

int calcBandD (Problem *problem) {
	int ans;
	// first - erase previous values of Di
	for (int i=0; i<problem->num_of_items; i++)
		Di[i] = 0;
	// fill out Di
	for (int j=0; j<problem->num_of_bids; j++)
		for (int i=0; i<problem->bids[j]->items_in_bid; i++)
			Di[problem->bids[j]->items[i]] += Yj[j];
	ans = calcB (problem);
	return ans;
}		

// computes unnormalized sum_y, and returns it, with no side-effects
double computeSum_Y(Problem *problem) {
	double result = 0;
	for( int j=0; j<problem->num_of_bids; j++ )
		result = result + Yj[j]* problem->bids[j]->offer;
	return result;
}

// returns true if the given bid num is rich
// a bid is defined as poor if its weight is less than 1/n^3 of the total sumY
bool richBid (Problem *problem, int bidNum) {
	return (Yj[bidNum] * problem->bids[bidNum]->offer > thresh);
}

// a bid is defined as poor if its weight is less than 1/n^3 of the total sumY
int richBids (Problem *problem) {
	int ans = 0;
	// for speed - return 0 ?M?
	return ans;

	for (int j=0; j<problem->num_of_bids; j++)
		if (richBid (problem,j))
			ans++;
	return ans;
}

// this is not simply a macro due to overflow problems - so need to convert
// to double, multiply by 1000 and divide by CLOCKS_PER_SEC
long clock2ms (clock_t t1, clock_t t2) {
	double diff = t2 - t1;
	diff *= 1000; // this step can cause overflow in long, so its done using double
	diff /= CLOCKS_PER_SEC;
	return (long)diff;
}
// used to compare 2 doubles, causing a descending order
int doubleCompare( const void *arg1, const void *arg2 )
{
   /* Compare all of both strings: */
	if (*(double *)arg1 < *(double *)arg2)
		return 1;
	if (*(double *)arg1 > *(double *)arg2)
		return -1;
	return 0;
}

// calculates what will be the value of D(b) after addition of delta
double nextValOfD (Problem *problem, int B) {
	double ans = 0;
	for (int j=1; j<=problem->YofX[B][0]; j++)
		ans += Yj[problem->YofX[B][j]]*exp(-delta/problem->bids[problem->YofX[B][j]]->offer);
	return ans;
}

// A good is maximal if maxD - D(good) < delta
int maximalGoods (Problem *problem) {
	int ans = 0;
	// for speed - return 0 ?M?
	return ans;
	double tmp = nextValOfD(problem,B);
	for (int i=0; i<problem->num_of_items; i++)
//		if (maxD - Di[i] < delta)
//		if (Di[i]/maxD > 0.1)
		if (Di[i] > tmp)
			ans++;
	return ans;
}

double computeSumY(Problem *problem);

double maxOffer (Problem *problem) {
	double ans = 0;
	for (int i=0; i<problem->num_of_bids; i++)
		if (ans < problem->bids[i]->offer)
			ans = problem->bids[i]->offer;
	return ans;
}

void addDelta () {
	Xi[B] += delta;
}

void calcAlphajFromScratch (Problem *problem) {
	// erase previous values of alphaj and fill out alphaj. alpha is minimum alphaj
	for (int j=0; j<problem->num_of_bids; j++) {
		alphaj[j] = 0;
		for (int i=0; i<problem->bids[j]->items_in_bid; i++)
			alphaj[j] += Xi[problem->bids[j]->items[i]];
	}

	for (int j=0; j<problem->num_of_bids; j++)
		alphaj[j] /= problem->bids[j]->offer;
	alpha = alphaj[0];
	for (int j=1; j<problem->num_of_bids; j++)
		if (alpha > alphaj[j])
			alpha = alphaj[j];
}

void calcAlphaj (Problem *problem, int B) {
	// only add to alphaj that changed due to Bi
	int bid;
	double dif;
	for (int j=1; j<=problem->YofX[B][0]; j++) {
		bid = problem->YofX[B][j];
		dif = Yj[bid];
		alphaj[bid] += delta/problem->bids[bid]->offer;
		Yj[bid] *= exp (-delta/problem->bids[bid]->offer);
//		dif -= Yj[bid];
		// now go over items i connected to current bid YofX[B][j] and update it 
		// according to new value of Yj
		for (int i=0; i<problem->bids[bid]->items_in_bid; i++) {
			Di[problem->bids[bid]->items[i]] += Yj[bid] - dif;
//			newY[problem->bids[bid]->items[i]] += Yj[bid];
//			oldY[problem->bids[bid]->items[i]] += dif;
		}
	}

//	// now go over Di's - and recalculate those with -1
//	for (int i=0; i<problem->num_of_items; i++) {
//		Di[i] -= oldY[i];
//		Di[i] += newY[i];
//	}
	// alpha is minimum alphaj
	alpha = alphaj[0];
	for (int j=1; j<problem->num_of_bids; j++)
		if (alpha > alphaj[j])
			alpha = alphaj[j];
}

// compute sum of x without any loops
// oldSum is the previous sum of x before normalizing
double computeSumX(double oldSum) {
	sum_x = oldSum + delta; // save un-normalized sum for next time
	return sum_x/alpha;
}

/*
	  Computes sumX.
	  sumX = Sum(i)Xi.
*/
double computeSumX(Problem *problem) {
	double result = 0;
	for( int i = 0; i < problem->num_of_items; i++ )
		result += Xi[i];
	return result/alpha;
}

/*
	  Compute sumY.
	  sumY = sum(j)Pj * Yj.
	  side effects: calculate sum_y and threshold for poor bids
*/
double computeSumY(Problem *problem) {
	sum_y = computeSum_Y (problem);
	thresh = sum_y / pow ((double)problem->num_of_bids,3.0);
	return sum_y/maxD;
}

void calcYj(Problem *problem, int sizeB, int Bi[]) {
	/*
	for (int i=0; i<sizeB; i++) 
		for (int j=1; j<=YofX[Bi[i]][0]; j++)
			Yj[YofX[Bi[i]][j]] = exp (-alphaj[YofX[Bi[i]][j]]);
			*/
}

void calcYj(Problem *problem) {
	for (int j=0; j<problem->num_of_bids; j++)
		Yj[j] = exp (-alphaj[j]);
}

void init (Problem *problem) {
	delta = 0;
	maxD = 0;
	if (Xi != NULL)
		free (Xi);
	Xi = new double[problem->num_of_items];
	// init xi with 0
	for (int i=0; i<problem->num_of_items; i++) 
		Xi[i] = 0;
	if (Yj != NULL)
		free (Yj);
	Yj = (double *)malloc (sizeof(double)*problem->num_of_bids);
	// init yj with 1, and delta with max price per bid
	for (int j=0; j<problem->num_of_bids; j++) {
		Yj[j] = 1;
		delta = max(delta,problem->bids[j]->offer);
	}
	if (Di != NULL)
		free (Di);
	Di = new double[problem->num_of_items];
	if (alphaj != NULL)
		free(alphaj);
	alphaj = (double *)malloc (sizeof(double)*problem->num_of_bids);
	for (int j=0; j<problem->num_of_bids; j++)
		alphaj[j] = 0;
	// now calculate YofX
    // later - realloc to size that you need. 0 place contains number of bids
	problem->YofX = (int **)malloc (sizeof(int *)*problem->num_of_items);
	if (!problem->YofX) {
		printf ("Unable to allocate enough memory for YofX");
		exit (1);
	}
	// at first - all products contain 0 bids
	for (int i=0; i<problem->num_of_items; i++) {
		problem->YofX[i] = (int *)malloc (sizeof(int)*(1)); // 18.4.01 try small size first
		problem->YofX[i][0] = 0;
	}
	int ii; // item index
	for (int j=0; j<problem->num_of_bids; j++)
		for (int i=0; i<problem->bids[j]->items_in_bid; i++) {
			ii = problem->bids[j]->items[i];
			problem->YofX[ii] = (int *)realloc(problem->YofX[ii],sizeof(int)*(2+problem->YofX[ii][0]));
			problem->YofX[ii][0]++;
			problem->YofX[ii][problem->YofX[ii][0]] = j;
		}
	// now release most of the memory...
	for (int i=0; i<problem->num_of_items; i++)
		problem->YofX[i] = (int *)realloc (problem->YofX[i],sizeof(int)*(problem->YofX[i][0]+1));
//	oldY = (double *)malloc (sizeof(double)*problem->num_of_items);
//	newY = (double *)malloc (sizeof(double)*problem->num_of_items);
}

double avgXinY(Problem *problem) {
	int sum = 0;
	for (int j=0; j<problem->num_of_bids; j++)
		sum += problem->bids[j]->items_in_bid;
	return (double)sum/(double)problem->num_of_bids;
}

int maxXinY(Problem *problem) {
	int ans=0;
	for (int j=0; j<problem->num_of_bids; j++)
		ans = max(ans,problem->bids[j]->items_in_bid);
	return ans;
}

double sumOffers (Problem *problem) {
	double ans = 0;
	for (int j=0; j<problem->num_of_bids; j++)
		ans += problem->bids[j]->offer;
	return ans;
}

/**
 * Returns weighted harmonic average of offers. = 
 * calc sum(Yj) where B in Bid j/ sum (Yj/offer(j)) where B in Bid j
 */
double weightedHarmonicAvg (int B, Problem *problem) {
	double nom = 0;
	double denom = 0;
	for (int j=1; j<=problem->YofX[B][0]; j++) {
		nom += Yj[problem->YofX[B][j]];
		denom += Yj[problem->YofX[B][j]]/problem->bids[problem->YofX[B][j]]->offer;
	}
	return (nom/denom);
}

double calcDelta (int B, double eps, Problem *problem) {
	return eps * weightedHarmonicAvg (B,problem);
}

Solution *solve (Problem *prob) {
	Problem *problem = prob;
	init (problem);
	double sumX = 0, sumY = 1.0;
	int counter = 0;
	clock_t start,finish,clockN;
	start = clock();
	finish = start;
	sum_x = 0;
	B = calcBandD (problem);
	sumY = computeSumY(problem);
	double eps = max(2,problem->epsilon);
	delta = calcDelta (B,eps,problem);
	// 18.4.01 - add printouts for really large problems - every 100 loops
	int printLoop = (problem->num_of_bids > 2000 ? 100 : problem->num_of_bids);
	bool cont = true;
	while (cont) {
		do {
			if (counter%printLoop == 0) {
				printf ("%d::delta: %e sumX: %f sumY: %f alpha: %f maxD: %e r: %f ",counter,delta,sumX,sumY,alpha,maxD, sumX/sumY-1);
				clockN = finish;
				finish = clock();
				printf ("MG: %d RB: %d ms: %d\n",maximalGoods(problem),richBids(problem),clock2ms(clockN,finish));
				fflush(stdout);
			}
			addDelta();
			calcAlphaj(problem,B);
			sumY = computeSumY(problem);
			sumX = computeSumX(sum_x);
			counter++;
			// calculate B for next iteration
			if (counter%50 == 0)
				B = calcBandD (problem); // for correcting small errors in maxD which occur 
										 // due to underflow
			else
				B = calcB(problem);
			delta = calcDelta(B,eps,problem);
		} while (alpha == 0 || sumX/sumY-1 > eps);
		if (eps <= problem->epsilon)
			cont = false;
		else 
			eps = max(problem->epsilon,eps*0.9);
	}
    printf("sumX: %f sumY: %f\n",sumX,sumY);
	finish = clock();
	Solution *solution = new Solution;
	solution->approxValue = sumX;
	for (int j=0; j<problem->num_of_bids; j++)
		Yj[j] /= maxD;
	
	solution->bidAllocation = Yj;
	for (int i=0; i<problem->num_of_items; i++)
		Xi[i] /= alpha;
	solution->productPrices = Xi;
	solution->elapsedTime = clock2ms(start,finish);
	free (Di); Di = NULL;
	free (alphaj); alphaj = NULL;
	printf ("Number of loops : %d\nAlgorithm calculation took: %d ms\n",counter,solution->elapsedTime);
	printf ("Returning from algorithm::solve with value %f\n",sumX);
	fflush (stdout);
	return solution;
}
