/**
 * Mar 31 : remove all use of array cancels
 * Apr 02 : add use of MAXYOFY so that it doesn't take up too much memory
 * Apr 02 : fix sort so that it doesn't create memory without freeing it.
 */
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include "GreedyValue.h"
#define MAXJ (prob->num_of_items) 
//#define MAXI (max(200,prob->num_of_items)) 18.4.01 - just for large problems
#define MAXI (max(200,min(400,prob->num_of_items)))
#define SMALL_BID_SIZE 0.2 // max percent of avg bid size vs number of items which allows YofY use
#define SMALL_NUM_BIDS 200 // max num of bids which still allow YofY use where avg bid size is large
#define MAXYOFY 2000	   // size of YofY is 4*(num_of_bids)^2, so 2000 bids -> 15 MB of memory
#define	eps 0.000001

int		*keys;
bool	*eligible = NULL;
Problem *prob;
Solution *sol;
double	*bidVals = NULL;
bool    *allocatedItems = NULL;
int		**YofY;

int *mergeForYofY (int sizeLeft, int *left, int sizeRight, int *right) {
	int lidx = 0,ridx = 0,idx=1;
    int *ans = (int *)malloc (sizeof(int)*(1+sizeLeft+sizeRight));
    while (lidx < sizeLeft && ridx < sizeRight) {
		if (left[lidx] == right[ridx]) {
			ans[idx] = left[lidx];
            lidx++;
			ridx++; // you can skip it, since it's already there
			idx++;
		} else if (left[lidx] < right[ridx]) {
			ans[idx] = left[lidx];
            lidx++;
			idx++;
        } else if (idx == 1 || ans[idx-1] < right[ridx]) {
			ans[idx] = right[ridx];
			ridx++;
			idx++;
		} else // otherwise this means that they are equal, so no need to add right[ridx]
			ridx++;
    }
    while (lidx < sizeLeft) {
        ans[idx] = left[lidx];
        lidx++;
		idx++;
    }
    while (ridx < sizeRight) {
		if (idx == 1 || ans[idx-1] < right[ridx]) {
			ans[idx] = right[ridx];
			ridx++;
			idx++;
		}
    }
	ans = (int *)realloc (ans,sizeof(int)*idx);
	ans[0] = idx-1;
    return ans;
}

void fillYofYofItem(int *bidNum) {
	int len = bidNum[0];
	int *tmp;
	if (len <= 1)
		return;
	// add all bid numbers in list to all bids in list, since they all share this item
	for (int i=1; i<=len; i++) {
		tmp = mergeForYofY(YofY[bidNum[i]][0],YofY[bidNum[i]]+1,len,bidNum+1);
		free (YofY[bidNum[i]]);
		YofY[bidNum[i]] = tmp;
	}
}

void fillYofY () {
	double avg = avgXinY(prob);
	if (avg > SMALL_BID_SIZE*prob->num_of_items && prob->num_of_bids > SMALL_NUM_BIDS) {
		printf ("average bid size is too large and num of bids too high so YofY isn't used\n");
		return;
	}
	if (prob->num_of_bids > MAXYOFY) {
		printf ("num of bids too high so YofY isn't used\n");
		return;
	}
	// now calculate YofX
    // later - realloc to size that you need. 0 place contains number of bids
	YofY = (int **)malloc (sizeof(int *)*prob->num_of_bids);
	if (!YofY) {
		printf ("Unable to allocate enough memory for YofY");
		exit (1);
	}
	// at first - all bids cancel 0 bids
	for (int i=0; i<prob->num_of_bids; i++) {
		YofY[i] = (int *)malloc (sizeof(int)*(1+prob->num_of_bids));
		YofY[i][0] = 0;
	}
	for (int i=0; i<prob->num_of_items; i++)
		fillYofYofItem (prob->YofX[i]);

	// now release most of the memory...
	for (int i=0; i<prob->num_of_bids; i++)
		YofY[i] = (int *)realloc (YofY[i],sizeof(int)*(YofY[i][0]+1));
}

void switchKeys (int *arg, int i1, int i2) {
	int tmp = arg[i1];
	arg[i1] = arg[i2];
	arg[i2] = tmp;
}

double bidVal(int idx) {
	double ans=0, sumPrice=0;
	for (int i=0; i<prob->bids[idx]->items_in_bid; i++)
        sumPrice += sol->productPrices[prob->bids[idx]->items[i]];
    ans = prob->bids[idx]->offer/sumPrice; // assuming no bid is empty ...
    return ans;
}

void calcBidVals () {
	bidVals = (double *)realloc (bidVals,sizeof(double)*prob->num_of_bids);
	for (int j=0; j<prob->num_of_bids; j++)
		if (sol->bidAllocation[j] <= eps)
			bidVals[j] = bidVal(j);
		else
			bidVals[j] = 0;
}

bool gt (int bid1, int bid2) {
	if (sol->bidAllocation[bid1] > eps && sol->bidAllocation[bid2] > eps)
        return (sol->bidAllocation[bid1] > sol->bidAllocation[bid2]);
    if (sol->bidAllocation[bid1] <= eps && sol->bidAllocation[bid2] <= eps)
        return (bidVals[bid1] > bidVals[bid2]);
    return (sol->bidAllocation[bid1] > eps);
}

double valueOfKeysUsingYofY(int *keys) {
	double ans = 0;
	if (eligible == NULL)
		eligible = new bool [prob->num_of_bids];
    for (int j=0; j<prob->num_of_bids; j++)
        eligible[j] = true;
	for (int j=0; j<prob->num_of_bids; j++) {
        if (!eligible[keys[j]])
          continue;
        eligible[keys[j]] = false;
        ans += prob->bids[keys[j]]->offer;
		// mark all bids in YofY[keys[j]] as not eligible
		for (int i=1; i<=YofY[keys[j]][0]; i++)
			eligible[YofY[keys[j]][i]] = false;
      }
	return ans;
}

double valueOfKeysOld(int *keys) {
	double ans = 0;
	if (eligible == NULL)
		eligible = new bool [prob->num_of_bids];
    for (int j=0; j<prob->num_of_bids; j++)
        eligible[j] = true;
	for (int j=0; j<prob->num_of_bids; j++) {
        if (!eligible[keys[j]])
          continue;
        eligible[keys[j]] = false;
        ans += prob->bids[keys[j]]->offer;
		for (int k=0; k<prob->num_of_bids; k++)
			if (eligible[k])
				for (int i1=0; i1<prob->bids[k]->items_in_bid; i1++) {
					if (!eligible[k])
						break;
					for (int i2=0; i2<prob->bids[keys[j]]->items_in_bid; i2++)
						if (prob->bids[k]->items[i1] == prob->bids[keys[j]]->items[i2]) {
							eligible[k] = false;
							break;
						}
				}
      }
	return ans;
}

// this time implement without eligible, but instead use allocatedItems
double valueOfKeys(int *keys) {
	double ans = 0;
	bool eligible = true;
	if (allocatedItems == NULL)
		allocatedItems = new bool [prob->num_of_items];
    for (int i=0; i<prob->num_of_items; i++)
        allocatedItems[i] = false;
	for (int j=0; j<prob->num_of_bids; j++) {
		// determine if bid is eligible - see if its items are allocated or not
		eligible = true;
		for (int i=0; i<prob->bids[keys[j]]->items_in_bid; i++)
			if (allocatedItems[prob->bids[keys[j]]->items[i]]) {
				eligible = false;
				break;
			}
        if (!eligible)
          continue;
		// now make all of current bid's items allocated
		for (int i=0; i<prob->bids[keys[j]]->items_in_bid; i++)
			allocatedItems[prob->bids[keys[j]]->items[i]] = true;
        ans += prob->bids[keys[j]]->offer;
      }
	return ans;
}

int *merge (int sizeLeft, int *left, int sizeRight, int *right) {
	int lidx = 0,ridx = 0;
    int *ans = (int *)malloc (sizeof(int) * (sizeLeft + sizeRight));
    while (lidx < sizeLeft && ridx < sizeRight) {
		if (gt (left[lidx],right[ridx])) {
			ans[lidx+ridx] = left[lidx];
            lidx++;
        } else {
            ans[lidx+ridx] = right[ridx];
            ridx++;
        }
    }
    while (lidx < sizeLeft) {
        ans[lidx+ridx] = left[lidx];
        lidx++;
    }
    while (ridx < sizeRight) {
        ans[lidx+ridx] = right[ridx];
        ridx++;
    }
    return ans;
}

int *sort (int size, int *arr) {
	if (size == 1)
		return arr;
	int med = size/2;
	int* left = sort (med,arr);
	int* right = sort (size-med,arr+med);
	int* ans = merge (med,left, size-med,right);
	if (med > 1)	  
		free (left);
	if (size-med > 1)	  
		free (right);
	return ans;
}


double oneJumpValue (int *arg) {
	int *local_keys = (int *)malloc(sizeof(int)*prob->num_of_bids);
	local_keys = (int *)memcpy (local_keys,arg,sizeof(int)*prob->num_of_bids);
	double ans = 0;
    double maxVal = (YofY == NULL) ? valueOfKeys (local_keys) : valueOfKeysUsingYofY (local_keys);
    double tmpVal = 0;
    int tmp;
	int maxi = min(prob->num_of_bids,MAXI);
	if (YofY == NULL) 
		for (int i=1; i<maxi; i++) {
			// switch first key with key i
			tmp = local_keys[0];
			local_keys[0] = local_keys[i];
			local_keys[i] = tmp;
			tmpVal = valueOfKeys(local_keys);
			if (tmpVal > maxVal)
				maxVal = tmpVal;
		}
	else 
		for (int i=1; i<maxi; i++) {
			// switch first key with key i
			tmp = local_keys[0];
			local_keys[0] = local_keys[i];
			local_keys[i] = tmp;
			tmpVal = valueOfKeysUsingYofY(local_keys);
			if (tmpVal > maxVal)
				maxVal = tmpVal;
		}
	ans = maxVal;
	free (local_keys);
	return ans;
}

double oneJumpValueIterative (int *arg, double minVal) {
	int *local_keys = (int *)malloc(sizeof(int)*prob->num_of_bids);
	local_keys = (int *)memcpy (local_keys,arg,sizeof(int)*prob->num_of_bids);
	double ans = 0;
    double maxVal = minVal;
    double tmpVal = 0;
    int maxLoop=0;
	int maxi = min(prob->num_of_bids,MAXI);
	bool improve;
	int cnt = 0;
	do {
		improve = false;
		printf ("phase : %d\tmax greedy so far: %f\n",cnt,maxVal);
		fflush (stdout);
		for (int i=1; i<maxi; i++) {
			// switch first key with key i
			switchKeys (local_keys,0,i);
			tmpVal = (YofY == NULL) ? valueOfKeys (local_keys) : valueOfKeysUsingYofY (local_keys);
			if (tmpVal > maxVal+eps) {
				maxVal = tmpVal;
				improve = true;
				break;
			}
		}
		cnt++;
	} while (improve);
	ans = maxVal;
	free (local_keys);
	return ans;
}

double twoJumpValueIterative (int *arg) {
	int *local_keys = (int *)malloc(sizeof(int)*prob->num_of_bids);
	local_keys = (int *)memcpy (local_keys,arg,sizeof(int)*prob->num_of_bids);
	double ans = 0;
    double maxVal = (YofY == NULL) ? valueOfKeys (local_keys) : valueOfKeysUsingYofY (local_keys);
	double tmpVal;
    int maxLoopi=0,maxLoopj=0;
	int maxi = min(prob->num_of_bids,MAXI);
	int maxj;
	bool improve;
	int cnt = 0,tmp;
	int saveJ=-1;
	if (YofY == NULL)
		do {
			improve = false;
			for (int i=0; i<maxi-1; i++) {
				switchKeys(local_keys,0,i);
				maxj = min(i+3,min(MAXJ,prob->num_of_bids));
				for (int j=1; j<maxj; saveJ = j++) {
					// switch keys i,j with key 0 and 1
					switchKeys(local_keys,1,j);
					tmpVal = valueOfKeys(local_keys);
					if (tmpVal > maxVal+eps) {
						maxVal = tmpVal;
						if (maxLoopi < i)
							maxLoopi = i;
						if (maxLoopj < j)
							maxLoopj = j;
						improve = true;
						printf ("%d ) %f found at (%d,%d)\n",cnt,maxVal,i,j);
						fflush (stdout);
						break;
					}
				}
				if (improve)
					break;
				else if (saveJ > 2) {
					// undo the changes made by j-switching
					tmp = local_keys[1];
					memmove (local_keys+1,local_keys+2,(saveJ-2)*sizeof(int));
					local_keys[saveJ-1] = tmp;
				}
			}
			cnt++;
		} while (improve);
	else
		do {
			improve = false;
			for (int i=0; i<maxi-1; i++) {
				switchKeys(local_keys,0,i);
				maxj = min(i+3,min(MAXJ,prob->num_of_bids));
				for (int j=2; j<maxj; saveJ = j++) {
					// switch keys i,j with key 0 and 1
					switchKeys(local_keys,1,j);
					tmpVal = valueOfKeysUsingYofY (local_keys);
					if (tmpVal > maxVal+eps) {
						maxVal = tmpVal;
						if (maxLoopi < i)
							maxLoopi = i;
						if (maxLoopj < j)
							maxLoopj = j;
						improve = true;
						printf ("%d ) %f found at (%d,%d)\n",cnt,maxVal,i,j);
						fflush (stdout);
						break;
					}
				}
				if (improve)
					break;
				else if (saveJ > 2) {
					// undo the changes made by j-switching
					tmp = local_keys[1];
					memmove (local_keys+1,local_keys+2,(saveJ-2)*sizeof(int));
					local_keys[saveJ-1] = tmp;
				}
			}
			cnt++;
		} while (improve);
	printf ("Max loops : %d,%d\n",maxLoopi,maxLoopj);
	fflush (stdout);
	ans = maxVal;
	free (local_keys);
	return ans;
}


// every improvement starts loop again
double localMaxValue (Problem *problem,Solution *solution) {
	printf ("localMaxValue starting...\n");
	fflush (stdout);
	double ans = 0;
	clock_t start,finish,t1,t2;
	long common;
	start = clock();
	prob = problem;
	sol = solution;
	calcBidVals();
	printf ("BidVals calculated...\n");
	fflush (stdout);
	// ** YofY
	
	t1 = clock();
	fillYofY();
	t2 = clock();
	printf ("fillYofY completed in %d ms\n",clock2ms(t1,t2));
	fflush (stdout);
	
	// ** keys
	keys = (int *)malloc (sizeof(int)*(problem->num_of_bids));
	for (int j=0; j<problem->num_of_bids; j++)
		keys[j] = j;
	int *tmpkeys = sort (problem->num_of_bids,keys);
	free (keys);
	keys = tmpkeys;

	printf ("keys sorted...\n");
	fflush (stdout);
	t1 = clock();
	common = clock2ms(start,t1); // this is the time common to all greedy methods
    double maxVal = (YofY == NULL) ? valueOfKeys (keys) : valueOfKeysUsingYofY (keys);
	t2 = clock();
	printf ("Regular greedy value: %f\nRegular greedy time %d ms\n",maxVal,common+clock2ms(t1,t2));
	fflush (stdout);
	double tmpVal = 0; // 18.4.01 - go straight to onejumpvalueiterative
/*
	t1 = clock();
    double tmpVal = oneJumpValue(keys);
	t2 = clock();
	printf ("1-Jump greedy value: %f\n1-Jump greedy time %d ms\n",tmpVal,common+clock2ms(t1,t2));
	fflush (stdout);
	maxVal = max(maxVal,tmpVal);
*/
	t1 = clock();
    tmpVal = oneJumpValueIterative(keys,0); // so I'll make the first jump
	t2 = clock();
	printf ("1-Jump Iterative greedy value: %f\n1-Jump Iterative greedy time %d ms\n",tmpVal,
		common+clock2ms(t1,t2));
	fflush (stdout);
	maxVal = max(maxVal,tmpVal);
	double saveVal = maxVal; // ?M? Edo 8.3.01 - so that 2-jump will start from 0
	maxVal = 0;
/*
	t1 = clock();
	maxVal = twoJumpValueIterative (keys);
	t2 = clock();
	printf ("2-Jump greedy value: %f\n2-Jump greedy time %d ms\n",maxVal,common+clock2ms(t1,t2));
	fflush (stdout);
*/
	finish = clock();
	printf ("localMaxValue took: %d ms\n",clock2ms(start,finish));
	ans = max(maxVal,saveVal);
	fflush (stdout);
	free (keys);
	free (eligible);
	if (YofY != NULL) {
		for (int i=0; i<problem->num_of_bids; i++)
			free (YofY[i]);
		free (YofY);
		YofY = NULL;
	}
	eligible = NULL;
	return ans;
}

