/*	ParseInput.cpp
 *	Edo Zurel 15/02/2001
 *	Used to parse devries input file into a bid structure
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "GreedyValue.h"

void usage () {
	printf("ALP.exe <file name> <epsilon>\n");
}

/*
 * Parse given file, creating a problem out of it
 */
Problem *parseFile (char *path) {
	char fname[256];
	strcpy (fname,path);
	printf ("reading %s ...\n",fname);
	fflush(stdout);
	FILE *inp = fopen (fname,"r");
	if (inp == NULL) {
		printf ("Problem reading file %s\nCheck path and name\n",fname);
		exit(1);
	}
	char line[8096];
	Problem *problem = new Problem;
	// first line: number of items
	problem->num_of_items = atoi(fgets(line,8096,inp));
	// second line: number of sets 
	problem->num_of_bids  = atoi(fgets(line,8096,inp));
	problem->bids = (Bid**)malloc(problem->num_of_bids*sizeof(Bid*));
	if (problem->bids == NULL) {
		printf("ParseInput::parseFile:Unable to allocate enough memory for the bids");
		exit(1);
	}
	printf ("created %d bids\n",problem->num_of_bids); fflush(stdout);

	/*
	each of the following lines contains: 
	bidamount for a set 
	number of items in this set 
	a (space separated) list of indices of items contained in this bidset 
	*/
	for (int i=0; i<problem->num_of_bids; i++) {
	        if (fgets(line,8096,inp) == NULL) {
		  printf("corrupt input: input file ended after %d bids\n",i);
		  exit(1);
		}
		problem->bids[i] = new Bid;
		problem->bids[i]->offer = atof(strtok(line," "));
		problem->bids[i]->items_in_bid = atoi(strtok(NULL," "));
		problem->bids[i]->items = new int[problem->bids[i]->items_in_bid];
		for (int j=0; j<problem->bids[i]->items_in_bid; j++)
			problem->bids[i]->items[j] = atoi(strtok(NULL," "));
	}
	fclose (inp);
	printf ("Problem has %d items and %d bids.\n",problem->num_of_items,problem->num_of_bids);
	fflush(stdout);
	return problem;
}

int main (int argc, char **argv) {
	Problem *problem = NULL;
	if (argc == 3) {
		double epsilon = atof (argv[2]);
		problem = parseFile (argv[1]);
		if (problem != NULL) {
			// solve it
			problem->epsilon = epsilon;
			Solution *solution = solve (problem);
			double val = localMaxValue (problem,solution);
			printf ("Greedy Value = %f\n",val);
			freeProblem (problem);
			problem = NULL;
		}
		
	} else 
		usage();
	return 0;
}
